const cluster = require("cluster");
const { By, until } = require('selenium-webdriver');
const webdriver = require('selenium-webdriver'); 
const helpers = require("./helpers");

require('chromedriver');

let workerPIDToBeKilled = ''; // current worker PID to be killed

if (cluster.isMaster) {
  const yargs = require("yargs/yargs");
  const { hideBin } = require("yargs/helpers");

  yargs(hideBin(process.argv))
    .command(
      "run [config] [testName]",
      "run the test",
      (yargs) => {
        return yargs.positional("config", {
          describe: "yaml file with test config",
        });
      },
      (argv) => {
        const yaml = require("js-yaml");
        const fs = require("fs");

        const config = yaml.load(fs.readFileSync(argv.config, "utf8"));

        // Calculate Load Distribution based on script weights
        const sumOfAllScriptWeights = config.scripts.reduce(
          (sum, s) => sum + s.weight,
          0
        );

        config.scripts = config.scripts.map((script, i) => {
          const startWeight = config.scripts
            .slice(0, i)
            .reduce((sum, s) => sum + s.weight, 0);

          const endWeight = startWeight + script.weight;
          const distrib = script.weight / sumOfAllScriptWeights;

          return {
            ...script,
            startWeight,
            endWeight,
            distrib,
          };
        });
      
        // Calculate when phases should start and end in seconds
        config.phases = config.phases.map((phase, i) => {
          const startTime = config.phases
            .slice(0, i)
            .reduce((sum, p) => sum + p.duration, 0);
          const endTime = startTime + phase.duration;

          const targetVusers = config.scripts.map((script,j)=>{
            return phase.rampTo * script.distrib;
          });

          return {
            ...phase,
            startTime,
            endTime,
            targetVusers,
          };
        });

        console.log("config", config);
        console.log("config.phases", config.phases);

        let currentVusers = 0; // This will keep track of how many vusers we have started at any point of time
        let initialVusers = config.scripts.map(()=>0); // The number of vusers started at the beginning of the current active phase
        let timeIndex = 0; // Elapsed time in seconds into the schedule we are
        let currentPhase = false; // The current phase of the config.phases we are in

        const workers = []; // an array we will add all the workers we spawn to

        const vusersPerScript = {};
        config.scripts.forEach(s=>vusersPerScript[s.processor]=0);

        let changeTargetVusersCalc = false; // To enable targetVusers calculation between ramp Up and ramp Down phases

        const publishToInflux = (currentVusers) => {
          helpers.writeActiveThreadsToInflux("virtualUsers", currentVusers);
        };

        // Execute every second
        setInterval(() => {
          // Get what phase we are in based on what time of the scenario we are in
          let phase = config.phases.find(
            (p) => p.startTime <= timeIndex && p.endTime > timeIndex
          );

          // if no phase test must be done
          if (!phase) {
            publishToInflux(currentVusers);
            console.log("At", timeIndex, "test has completed!");
            process.exit(0);
          }

          // If we have started a new phase of the test
          if (phase != currentPhase) {
            console.log("At", timeIndex, "Started phase", phase);

            // Change targetVusers calculation while ramping down
            // Do not change targetVusers calculation while ramping up
            if(phase.rampTo < currentPhase.rampTo){
              changeTargetVusersCalc = true;
            } else {
              changeTargetVusersCalc = false;
            }

            currentPhase = phase;
            initialVusers = config.scripts.map(s=>vusersPerScript[s.processor] || 0);
          }

          // Loop through each script and determine targetVusers for this time differently for each script
          config.scripts.forEach((selectedScript,scriptIndex) => {
            const scriptRampTo = currentPhase.targetVusers[scriptIndex];
            const initialVusersForScript = initialVusers[scriptIndex];
            let currentVusersForScript = vusersPerScript[selectedScript.processor];
              
            // Calculate how many vusers should be running right now
            const perc = (timeIndex - phase.startTime) / phase.duration;

            // Change targetVusers calculation technique based on rampUp/rampDown load phase
            let targetVusers = 0; // The number of vusers we are expected to be at right now for this script
            if(changeTargetVusersCalc) { // only called during ramp down
              targetVusers = Math.ceil((scriptRampTo - initialVusersForScript - 1) * perc) + initialVusersForScript
            } else{ // only called during ramp up or steady
              targetVusers = Math.ceil((scriptRampTo - initialVusersForScript) * perc) + initialVusersForScript
            }

            if (targetVusers > currentVusersForScript) {
              // Start more worker threads

              for (let i = currentVusersForScript; i < targetVusers; i++) {

                if(typeof vusersPerScript[selectedScript.processor]=="undefined")
                          vusersPerScript[selectedScript.processor]=0;
                
                const newWorker = cluster.fork({
                  SCRIPT_NAME: selectedScript.processor,
                  threadNum: currentVusers,
                  vuserNum: vusersPerScript[selectedScript.processor],
                  SELENIUM_HOST: process.env.SELENIUM_HOST,
                });


                // To remove limit to number of Worker listeners
                newWorker.setMaxListeners(0);
                newWorker.scriptProcessor = selectedScript.processor;

                console.log("Worker", newWorker.process.pid, "started");

                workers.push(newWorker);

                vusersPerScript[selectedScript.processor]++;
                currentVusers++;
                currentVusersForScript++;

                publishToInflux(currentVusers);
              }
            } else if (targetVusers < currentVusersForScript) {
              // End running worker threads
              let worker = workers.find(w=>w.scriptProcessor==selectedScript.processor);
              if(!worker) return;
                // Kill a new child process when it is ready
                if(workerPIDToBeKilled !== worker.process.pid) {
                  // Set workerPIDToBeKilled to the current worker pid
                  workerPIDToBeKilled = worker.process.pid;						
                  // Send to Child process the PID about to be killed
                  worker.send(workerPIDToBeKilled);
                  // Listen for 'Kill' message from Child process before killing
                  worker.on("message", (msg) => {
                    try {
                      if(msg === 'Kill') {
                        worker.kill();
                        currentVusers--;
                        currentVusersForScript--;
                        vusersPerScript[selectedScript.processor]--;
                        workers.shift();
                      }
                    } catch(e){
                      console.log(e);
                    }
                  });
                }

              publishToInflux(currentVusers);
            } else {
              publishToInflux(currentVusers);
            }
          });

          timeIndex++;
        }, 1000);

        cluster.on("exit", (worker, code, signal) => {
          if (signal) {
            console.log(`worker was killed by signal: ${signal}`);
          } else if (code !== 0) {
            console.log(`worker exited with error code: ${code}`);
          } else {
            console.log("worker success!");
          }
          publishToInflux(currentVusers);
        });
      }
    )
    .strictCommands()
    .demandCommand(1)
    .parse();
} else {
    // Increment the threadNum number for the child process to start from threadNum=1
    process.env.threadNum++;

    const path = require("path");
    const fname = path.join(process.cwd(), process.env.SCRIPT_NAME);
    console.log(
      "Im a child worker!",
      process.pid,
      "that will execute:",
      fname,
      "thread number: ",
      process.env.threadNum
    );

    const script = require(fname);

    // Initialize Chrome Driver

    let cap = webdriver.Capabilities.chrome();
    let chromeOptions = {
      args: ["--headless", "--start-maximized","--no-sandbox", "--disable-site-isolation-trials","--renderer-process-limit=1","--in-process-gpu","--no-zygote", "--disable-extensions", "--log-level=3"],
      excludeSwitches: ["enable-automation"],
      useAutomationExtension: false,
    };

    cap.set("goog:chromeOptions", chromeOptions);
    cap.set("acceptSslCerts", true);
    cap.set("acceptInsecuerts", true);

    let driver = new webdriver.Builder()
      .forBrowser("chrome")
      .withCapabilities(cap)
      .usingServer(`http://${process.env.SELENIUM_HOST}/wd/hub`)
      .build();

    // execute the script over and over until this child process is killed
    const params = {
      driver,
      threadNum: process.env.threadNum,
      vuserNum: process.env.vuserNum,
      By,
      until,
      helpers,
      webdriver
    };

    // To store the received PID to be killed
    let processPID = '';

    
      // To ensure that this is a test engine execution (no creating / writing to files performed)
      helpers.setIsEngine(true);

      // To remove limit to number of Process listeners
      process.setMaxListeners(0);

    // execute the script over and over until this child process is killed
    (async () => {


      // The init script should only run once
      if (script.init) {
        try {
          await script.init(params);
        } catch(err) {
          // Write to console the error
          console.error(`error occured during vuser_init. vuser = ${params.threadNum}`,err);
          // Exit the process
          process.exit(0);
        }
      }

      while (true) {
        try { 
            // Receive the Process PID from Master
            process.on("message", (msg) => {
              processPID = msg;
            });

            // If received PID is the same as the currently running PID, then run script.end and send Kill signal back to Master
            // Otherwise, continue running script.main
            if(processPID === process.pid) {
              console.log(`Process ${process.pid} about to run script.end`);
                if(script.end) {
                  try {
                    await script.end(params); 
                  } catch(err) {
                    console.error(`error occured during vuser_end. vuser = ${params.threadNum}`,err);
                  } finally {
                    return process.send('Kill');
                  }
                }
            } else {
              await script.main(params);
            }
            
        } catch (err) {
            console.error(`vuser = ${params.threadNum}, had an uncaught error`, err);
            const endTransaction = helpers.endTransactionFactory(params.threadNum, driver);
            helpers.openTransactions.forEach(name => {
              endTransaction(name, false);
            });
        }
      }
    })();
}
